#include "../../std.h"

#include <iostream>
#include <fstream>

#include "../../misc/stringtoint.h"
#include "../../misc/escaped.h"
#include "../../misc/time_utils.h"
#include "files.h"


SEFileHandle::SEFileHandle(SEFile& f,uint64_t o,bool for_read):file(f) {
  active=true;
  offset=o;
  open_read=for_read;
  if(file.open(open_read) != 0) active=false;
cerr<<"SEFileHandle constructor: active: "<<active<<endl;
}

SEFileHandle::~SEFileHandle(void) {
  if(active) file.close(open_read);
}

uint64_t SEFileHandle::read(void* buf,uint64_t size) {
cerr<<"SEFileHandle::read - size: "<<size<<endl;
  if(!active) return 0;
  uint64_t l = file.read(buf,offset,size);
  offset+=l;
cerr<<"SEFileHandle::read - l: "<<l<<endl;
  return l;
}

uint64_t SEFileHandle::write(void* buf,uint64_t size) {
  if(!active) return 0;
  uint64_t l = file.write(buf,offset,size);
  offset+=l;
  return l;
}

SEFileHandle* SEFileHandle::open(const char* id,uint64_t o,bool for_read,SEFiles& files) {
  SEFile* file = files.get(id);
  if(!file) return NULL;
  SEFileHandle* f = new SEFileHandle(*file,o,for_read);
  if(!(*f)) { delete f; return NULL; };
  return f;
}

SEFiles::SEFiles(void) {
cerr<<"SEFiles::SEFiles(void)"<<endl;
//  files=NULL;
  ns=NULL;
//  nfiles=0;
  valid=false;
  reg_type=registration_immediate;
}

SEFiles::SEFiles(const char* dirpath):path(dirpath) {
cerr<<"SEFiles::SEFiles(const char* dirpath)"<<endl;
  valid=false;
//  files=NULL;
  ns=NULL;
  reg_type=registration_immediate;
//  nfiles=0;
  struct dirent file_;
  struct dirent *file;
  DIR *dir=opendir(dirpath);
  if(dir == NULL) {
    cerr<<"Failed opening directory: "<<dirpath<<endl; return;
  };
  for(;;) {
    readdir_r(dir,&file_,&file);
    if(file == NULL) break;
    int l=strlen(file->d_name);
    if(l>5) {
      if(!strcmp(file->d_name+l-5,".attr")) {
        string name(file->d_name); name.resize(l-5);
        name="/"+name; name=dirpath+name;
cerr<<"SEFiles: creating SEFile: "<<name<<endl;
        SEFile* f = new SEFile(name.c_str());
cerr<<"SEFiles: adding SEFile: "<<name<<endl;
        add(*f);
cerr<<"Added file: "<<f->id()<<endl;
      };
    };
  };
  closedir(dir);
  valid=true;
}

SEFiles::~SEFiles() {
  lock.block();
//  for(int i=0;i<nfiles;i++) {
//    if(files[i]) delete files[i];
//  };
//  free(files); files=NULL;
  valid=false;
  lock.unblock();
}

SEFile* SEFiles::add(SEFile& file) {
cerr<<"SEFiles::add"<<endl;
  lock.block(); // TODO: to remove ?!
  if(files) {
    SEFile* f = find_by_id(file.id());
    if(f) {
      // File with this id already exists. Allow (re)creation if 
      // owner is a same person (DN) and other attributes are same
      // too (if set).
      bool same = true;
      if(file.creator() != f->creator()) same=false;
      if(same && file.size_available() && f->size_available()) {
        if(file.size() != f->size()) same=false;
      };
      if(same && file.checksum_available() && f->checksum_available()) {
        if(file.checksum() != f->checksum()) same=false;
      };
      if(same && file.created_available() && f->created_available()) {
        if(memcmp(file.created(),f->created(),sizeof(struct tm))!=0)same=false;
      };
      if(!same) {
        cerr<<"SEFiles::add: file already exists: "<<file.id()<<endl;
        file.destroy();
        lock.unblock();
        return NULL;
      };
      // files look identical, destroy new and replace it by new
      file.destroy();
      lock.unblock();
      return f;
    };
  };
  cerr<<"SEFiles::add: new file: "<<file.id()<<endl;
  files.add(file);
  lock.unblock();
  return &file;
}

bool SEFiles::remove(SEFile& file) {
  for(SEFiles::iterator i=begin();i!=end();++i) {
    if(&(*i) == &file) {
      return remove(i);
    };
  };
  return false;
}

bool SEFiles::remove(SEFiles::iterator& file) {
cerr<<"SEFiles::remove: "<<file->id()<<endl;
  if(!file) return true;
  lock.block(); // TODO: to remove
cerr<<"SEFiles::remove: calling destroy"<<endl;
  file->destroy(); 
  files.remove(file);
  lock.unblock();
  return true;
}

SEFile* SEFiles::find_by_id(const char* id) {
cerr<<"find_by_id: id: "<<id<<endl;
  for(SEFiles::iterator i = begin();i!=end();++i) {
cerr<<"find_by_id: i: "<<i->id()<<endl;
    if(i->check_id(id)) {
      SEFile* f = &(*i);
      return f;
    };
  };
/*
  for(int i=0;i<nfiles;i++) {
    if(files[i]) if(files[i]->check_id(id)) {
      SEFile* f = files[i];
      return f;
    };
  };
*/
  return NULL;
}

SEFile* SEFiles::get(const char* id) {
  lock.block(); // TODO: to remove
  SEFile* f = find_by_id(id);
  lock.unblock();
  return f;
}

// Register all files, which require registration
int SEFiles::Register(void) {
cerr<<"SEFiles::Register"<<endl;
  // TODO: Be careful with NS - it can be non-thread-safe
  int failures = 0;
  if(files && ns) {
//cerr<<"SEFiles::Register: have both files and NS"<<endl;
    for(SEFiles::iterator f = files.begin();f!=files.end();++f) {
//cerr<<"SEFiles::Register: have valid file: "<<f->id()<<endl;
      ns->Connect(); // take advantage of possibly less reconnections
      // analyze situation
      f->acquire();
//cerr<<"SEFiles::Register: state: "<<f->state_file()<<endl;
//cerr<<"SEFiles::Register: reg state: "<<f->state_reg()<<endl;
//cerr<<"SEFiles::Register: retry: "<<register_retry()<<endl;
//cerr<<"SEFiles::Register: immediate: "<<register_immediately()<<endl;
      if((f->state_reg() == REG_STATE_UNREGISTERING) ||
         (f->state_reg() == REG_STATE_REGISTERING)) {
        f->release(); continue; // file is alredy claimed
      };
      if(!register_retry()) { // this function only serves retries
        f->release(); continue;
      };
      if((f->state_file()!=FILE_STATE_COMPLETE) &&
         !(register_immediately()) && (f->state_file()!=FILE_STATE_COLLECTING)
        ) { // not proper for register
        f->release(); continue;
      };
      if(f->state_reg()!=REG_STATE_LOCAL) {  // already/yet registered 
        f->release(); continue;
      };
      // claim file being registered
cerr<<"Registering: "<<f->id()<<endl;
      if(!f->state_reg(REG_STATE_REGISTERING)) {
cerr<<"Registering: "<<f->id()<<" - change state failed"<<endl;
        failures++;
        f->release(); continue;
      };
      f->release();
      if(ns->Register(*f) != 0) {
        failures++;
        if(!f->state_reg(REG_STATE_LOCAL)) {
          // ???????????
        };
        continue;
      };
      if(f->state_reg(REG_STATE_ANNOUNCED)) {
        // ???????????
        continue;
      };
    };
    ns->Disconnect();
  };
  return failures;
}

// Unregister all files being deleted
int SEFiles::Unregister(void) {
cerr<<"SEFiles::Unregister"<<endl;
  int failures = 0;
  if(files && ns) {
//cerr<<"SEFiles::Unregister: have both files and NS"<<endl;
    for(SEFiles::iterator f = files.begin();f!=files.end();++f) {
//cerr<<"SEFiles::Unregister: have valid file: "<<f->id()<<endl;
      ns->Connect(); // take advantage of possibly less reconnections
      // analyze situation
      f->acquire(); // lock file to prevent race while analyzing state
//cerr<<"SEFiles::Unregister: state: "<<f->state_file()<<endl;
//cerr<<"SEFiles::Unregister: reg state: "<<f->state_reg()<<endl;
//cerr<<"SEFiles::Unregister: retry: "<<register_retry()<<endl;
//cerr<<"SEFiles::Unregister: immediate: "<<register_immediately()<<endl;
      if((f->state_reg() == REG_STATE_UNREGISTERING) ||
         (f->state_reg() == REG_STATE_REGISTERING)) {
        f->release(); continue; // file is already claimed
      };
      if(!register_retry()) { // this function only serves retries
        f->release(); continue;
      };
      if(f->state_file()!=FILE_STATE_DELETING) { // not proper for unregister
        f->release(); continue;
      };
      if(f->state_reg()==REG_STATE_LOCAL) {  // stuck file 
        f->release(); f->destroy(); files.remove(f); continue;
      };
      if(f->state_reg()!=REG_STATE_ANNOUNCED) {  // already/yet not registered 
        f->release(); continue;
      };
      // claim file being unregistered
cerr<<"Unregistering: "<<f->id()<<endl;
      if(!f->state_reg(REG_STATE_UNREGISTERING)) {
cerr<<"Unregistering: "<<f->id()<<" - change state failed"<<endl;
        failures++;
        f->release(); continue;
      };
      f->release();
cerr<<"Unregistering - calling Unregister"<<endl;
      if(ns->Unregister(*f) != 0) {
cerr<<"Unregistering - Unregister failed"<<endl;
        failures++;
        if(!f->state_reg(REG_STATE_ANNOUNCED)) {
          // ???????????
        };
        continue;
      };
cerr<<"Unregistering - Unregister succeeded"<<endl;
      if(!f->state_reg(REG_STATE_LOCAL)) {
        failures++;
        // ???????????
        continue;
      };
      // unregistered file in deleting state  - destroy and remove
      f->destroy();
      files.remove(f);
    };
    ns->Disconnect();
  };
  return failures;
}

int SEFiles::RemoveStuck(void) {
cerr<<"SEFiles::RemoveStuck"<<endl;
  int failures = 0;
  if(files) {
//cerr<<"SEFiles::RemoveStuck: have files"<<endl;
    for(SEFiles::iterator f = files.begin();f!=files.end();++f) {
//cerr<<"SEFiles::RemoveStuck: have valid file: "<<f->id()<<endl;
      // analyze situation
      f->acquire();
//cerr<<"SEFiles::RemoveStuck: state: "<<f->state_file()<<endl;
      if(f->state_file() == FILE_STATE_COLLECTING) {
        if((time(NULL) - f->state_file_changed()) > (24*60*60)) {
          if((time(NULL) - f->data_changed()) > (12*60*60)) {
            // file was not touched for long time - probably forgotten
            cerr<<"Removing file which stayed too long in COLLECTING state: "<<f->id()<<endl;
            bool unregistered = true;
            if(f->state_reg() == REG_STATE_ANNOUNCED) {
              unregistered=false;
              if(!f->state_reg(REG_STATE_UNREGISTERING)) {
                cerr<<"SEFiles::RemoveStuck: failed to set REG_STATE_UNREGISTERING"<<endl;
              } else {
                f->release();
                if(NS()->Unregister(*f) != 0) {
                  f->acquire();
                  f->state_reg(REG_STATE_ANNOUNCED);
                  cerr<<"SEFiles::RemoveStuck: failed to unregister"<<endl;
                } else {
                  f->acquire();
                  f->state_reg(REG_STATE_LOCAL);
                  unregistered=true;
                };
              };
            };
            if(unregistered) {
              remove(*f);
            };
          };
        };
      };
      f->release();
    };
  };
  return failures;
}

int SEFiles::Replicate(void) {
cerr<<"SEFiles::Replicate"<<endl;
  int failures = 0;
  if(files) {
//cerr<<"SEFiles::Replicate: have files"<<endl;
    for(SEFiles::iterator f = files.begin();f!=files.end();++f) {
//cerr<<"SEFiles::Replicate: have valid file: "<<f->id()<<endl;
      // analyze situation
      f->acquire();
//cerr<<"SEFiles::Replicate: state: "<<f->state_file()<<endl;
      if(f->state_file() == FILE_STATE_REQUESTED) {
        // claim file being downloaded
cerr<<"Downloading: "<<f->id()<<endl;
        if(!f->state_file(FILE_STATE_DOWNLOADING)) {
cerr<<"Downloading: "<<f->id()<<" - change state failed"<<endl;
          failures++; f->release(); continue;
        };
        f->release();
        bool downloaded = true;
        SEReplicator replicator;
cerr<<"Downloading - calling Obtain"<<endl;
        if(replicator.Obtain(*f) != 0) {
cerr<<"Downloading - Obtain failed"<<endl;
          failures++; downloaded=false;
        };
        f->acquire();
        if(downloaded) {
          if(!f->state_file(FILE_STATE_COMPLETE)) {
cerr<<"Downloading: "<<f->id()<<" - change state failed"<<endl;
            failures++; f->release(); continue;
          };
          // TODO: registration


        } else {
          if(!f->state_file(FILE_STATE_REQUESTED)) {
cerr<<"Downloading: "<<f->id()<<" - change state failed"<<endl;
            failures++; f->release(); continue;
          };
        };
      };
      f->release();
    };
  };
  return failures;
}

