#include "globus_modules.h"

#include <globus_common.h>
#include <globus_io.h>
#include <globus_gass_transfer.h>
#include <globus_gass_copy.h>
#include <globus_ftp_control.h>
#include <globus_ftp_client.h>
#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
#include <globus_replica_catalog.h>
#endif
#include <globus_gsi_credential.h>
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
#include <globus_rls_client.h>
#endif

globus_mutex_t GlobusModuleGlobalLock::mutex = PTHREAD_MUTEX_INITIALIZER;

void GlobusModuleGlobalLock::lock() {
  pthread_mutex_lock(&mutex);
}

void GlobusModuleGlobalLock::unlock() { 
  pthread_mutex_unlock(&mutex);
}

#define NEW_GLOBUS_MODULE_CLASS(class_name,module_name,class_string) \
unsigned int class_name::counter = 0; \
class_name::class_name(void) { \
  activated=activate(); \
}; \
class_name::~class_name(void) { \
  if(activated) deactivate(); \
}; \
bool class_name::activate(void) { \
  lock(); \
  int res = GLOBUS_SUCCESS; \
  if(counter==0) res=globus_module_activate(module_name); \
  if(res==GLOBUS_SUCCESS) counter++; \
  unlock(); \
  return (res==GLOBUS_SUCCESS); \
}; \
bool class_name::deactivate(void) { \
  lock(); \
  int res = GLOBUS_SUCCESS; \
  if(counter>0) { \
    counter--; \
    if(counter==0) res=globus_module_deactivate(module_name); \
    if(res != GLOBUS_SUCCESS) counter++; \
  }; \
  unlock(); \
  return (res==GLOBUS_SUCCESS); \
}

NEW_GLOBUS_MODULE_CLASS(GlobusModuleCommon,GLOBUS_COMMON_MODULE,"COMMON")

NEW_GLOBUS_MODULE_CLASS(GlobusModuleGASSTransfer,GLOBUS_GASS_TRANSFER_MODULE,"GASS_TRANSFER")

/*
NEW_GLOBUS_MODULE_CLASS(GlobusModuleGASSCopy,GLOBUS_GASS_COPY_MODULE,"GASS_COPY")
*/

NEW_GLOBUS_MODULE_CLASS(GlobusModuleFTPControl,GLOBUS_FTP_CONTROL_MODULE,"FTP CONTROL")

NEW_GLOBUS_MODULE_CLASS(GlobusModuleFTPClient,GLOBUS_FTP_CLIENT_MODULE,"FTP_CLIENT")

#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
NEW_GLOBUS_MODULE_CLASS(GlobusModuleReplicaCatalog,GLOBUS_REPLICA_CATALOG_MODULE,"REPLICA CATALOG")
#endif

NEW_GLOBUS_MODULE_CLASS(GlobusModuleGSIGSSAPI,GLOBUS_GSI_GSSAPI_MODULE,"GSI GSSAPI")

NEW_GLOBUS_MODULE_CLASS(GlobusModuleError,GLOBUS_ERROR_MODULE,"ERROR")

NEW_GLOBUS_MODULE_CLASS(GlobusModuleIO,GLOBUS_IO_MODULE,"IO")

/*
NEW_GLOBUS_MODULE_CLASS(GlobusModuleRSL,GLOBUS_RSL_MODULE,"RSL")
*/

NEW_GLOBUS_MODULE_CLASS(GlobusModuleGSICredential,GLOBUS_GSI_CREDENTIAL_MODULE,"GSI CREDENTIAL")

#ifdef HAVE_GLOBUS_RLS_CLIENT_H
NEW_GLOBUS_MODULE_CLASS(GlobusModuleRLSClient,GLOBUS_RLS_CLIENT_MODULE,"RLS");
#endif

